import ChartNS from 'chart.js';
import { forceSimulation, forceCenter, forceCollide, forceLink, forceManyBody, forceX, forceY, forceRadial } from 'd3-force';
import { hierarchy, tree, cluster } from 'd3-hierarchy';

const Chart = ChartNS;
// export const plugins = ChartNS.plugins;
const defaults = ChartNS.defaults;

// export const Scale = ChartNS.Scale;
// export const LinearScale = ChartNS.scaleService.getScaleConstructor('linear');
// export const LogarithmicScale = ChartNS.scaleService.getScaleConstructor('logarithmic');

// export const DatasetController = ChartNS.DatasetController;
// export const BarController = ChartNS.controllers.bar;
// export const BubbleController = ChartNS.controllers.bubble;
// export const HorizontalBarController = ChartNS.controllers.horizontalBar;
// export const LineController = ChartNS.controllers.line;
// export const PolarAreaController = ChartNS.controllers.polarArea;
const ScatterController = ChartNS.controllers.scatter;

function registerController(controller) {
  ChartNS.controllers[controller.id] = controller;
  ChartNS.defaults.set(controller.id, controller.defaults);
  return controller;
}

// export const Element = ChartNS.Element;
// export const Rectangle = ChartNS.elements.Rectangle;
// export const Point = ChartNS.elements.Point;
const Line = ChartNS.elements.Line;
// export const Arc = ChartNS.elements.Arc;

function registerElement(element) {
  ChartNS.defaults.set('elements', {
    [element.id]: element.defaults,
  });
  return element;
}

const merge = ChartNS.helpers.merge;
// export const drawPoint = ChartNS.helpers.canvas.drawPoint;
// export const resolve = ChartNS.helpers.options.resolve;
// export const color = ChartNS.helpers.color;
// export const valueOrDefault = ChartNS.helpers.valueOrDefault;
const clipArea = ChartNS.helpers.canvas.clipArea;
const unclipArea = ChartNS.helpers.canvas.unclipArea;

function patchControllerConfig(config, controller) {
  controller.register();
  config.type = controller.id;
  return config;
}

class EdgeLine extends Line {
  draw(ctx) {
    const options = this.options;

    ctx.save();

    // Stroke Line Options
    ctx.lineCap = options.borderCapStyle;
    ctx.setLineDash(options.borderDash || []);
    ctx.lineDashOffset = options.borderDashOffsetborderDash;
    ctx.lineJoin = options.borderJoinStyle;
    ctx.lineWidth = options.borderWidth;
    ctx.strokeStyle = options.borderColor;

    // Stroke Line
    ctx.beginPath();

    const from = this.source.getProps(['x', 'y', 'angle']);
    const to = this.target.getProps(['x', 'y', 'angle']);
    const angleHelper = Math.hypot(to.x - from.x, to.y - from.y) * options.tension;

    const orientations = {
      horizontal: {
        fx: (to.x - from.x) * options.tension,
        fy: 0,
        tx: (from.x - to.x) * options.tension,
        ty: 0,
      },
      vertical: {
        fx: 0,
        fy: (to.y - from.y) * options.tension,
        tx: 0,
        ty: (from.y - to.y) * options.tension,
      },
      radial: {
        fx: Number.isNaN(from.angle) ? 0 : Math.cos(from.angle) * angleHelper,
        fy: Number.isNaN(from.angle) ? 0 : Math.sin(from.angle) * -angleHelper,
        tx: Number.isNaN(to.angle) ? 0 : Math.cos(to.angle) * -angleHelper,
        ty: Number.isNaN(to.angle) ? 0 : Math.sin(to.angle) * angleHelper,
      },
    };
    const shift = orientations[this._orientation] || orientations.horizontal;

    const fromX = {
      cpx: from.x + shift.fx,
      cpy: from.y + shift.fy,
    };
    const toX = {
      cpx: to.x + shift.tx,
      cpy: to.y + shift.ty,
    };

    ctx.moveTo(from.x, from.y);
    // Line to next point
    if (options.stepped === 'middle') {
      const midpoint = (from.x + to.x) / 2.0;
      ctx.lineTo(midpoint, from.y);
      ctx.lineTo(midpoint, to.y);
      ctx.lineTo(to.x, to.y);
    } else if (options.stepped === 'after') {
      ctx.lineTo(from.x, to.y);
      ctx.lineTo(to.x, to.y);
    } else if (options.stepped) {
      ctx.lineTo(to.x, from.y);
      ctx.lineTo(to.x, to.y);
    } else if (options.tension) {
      ctx.bezierCurveTo(fromX.cpx, fromX.cpy, toX.cpx, toX.cpy, to.x, to.y);
    } else {
      ctx.lineTo(to.x, to.y);
    }

    ctx.stroke();
    ctx.restore();

    // point helper
    // ctx.save();
    // ctx.strokeStyle = 'blue';
    // ctx.beginPath();
    // ctx.moveTo(from.x, from.y);
    // ctx.lineTo(from.x + shift.fx, from.y + shift.fy, 3, 3);
    // ctx.stroke();
    // ctx.strokeStyle = 'red';
    // ctx.beginPath();
    // ctx.moveTo(to.x, to.y);
    // ctx.lineTo(to.x + shift.tx, to.y + shift.ty, 3, 3);
    // ctx.stroke();
    // ctx.restore();
  }
}

EdgeLine.id = EdgeLine._type = 'edgeLine';
EdgeLine.defaults = Object.assign({}, defaults.elements.line, {
  tension: 0,
});
EdgeLine.register = () => registerElement(EdgeLine);

const arrayEvents = ['push', 'pop', 'shift', 'splice', 'unshift'];

// COPIED from chart.js since not exposed

/**
 * Hooks the array methods that add or remove values ('push', pop', 'shift', 'splice',
 * 'unshift') and notify the listener AFTER the array has been altered. Listeners are
 * called on the 'onData*' callbacks (e.g. onDataPush, etc.) with same arguments.
 */
function listenArrayEvents(array, listener) {
  if (array._chartjs) {
    array._chartjs.listeners.push(listener);
    return;
  }

  Object.defineProperty(array, '_chartjs', {
    configurable: true,
    enumerable: false,
    value: {
      listeners: [listener],
    },
  });

  arrayEvents.forEach((key) => {
    const method = 'onData' + key.charAt(0).toUpperCase() + key.slice(1);
    const base = array[key];

    Object.defineProperty(array, key, {
      configurable: true,
      enumerable: false,
      value() {
        const args = Array.prototype.slice.call(arguments);
        const res = base.apply(this, args);

        array._chartjs.listeners.forEach((object) => {
          if (typeof object[method] === 'function') {
            object[method].apply(object, args);
          }
        });

        return res;
      },
    });
  });
}

/**
 * Removes the given array event listener and cleanup extra attached properties (such as
 * the _chartjs stub and overridden methods) if array doesn't have any more listeners.
 */
function unlistenArrayEvents(array, listener) {
  const stub = array._chartjs;
  if (!stub) {
    return;
  }

  const listeners = stub.listeners;
  const index = listeners.indexOf(listener);
  if (index !== -1) {
    listeners.splice(index, 1);
  }

  if (listeners.length > 0) {
    return;
  }

  arrayEvents.forEach((key) => {
    delete array[key];
  });

  delete array._chartjs;
}

class GraphController extends ScatterController {
  constructor(chart, datasetIndex) {
    super(chart, datasetIndex);

    this._initialReset = true;
    this._cachedEdgeOpts = {};
    this._edgeListener = {
      onDataPush: (...args) => {
        const count = args.length;
        const start = this.getDataset().edges.length - count;
        const parsed = this._cachedMeta._parsedEdges;
        args.forEach((edge) => {
          parsed.push(this._parseDefinedEdge(edge));
        });
        this._insertEdgeElements(start, count);
      },
      onDataPop: () => {
        this._cachedMeta.edges.pop();
        this._cachedMeta._parsedEdges.pop();
        this._scheduleResyncLayout();
      },
      onDataShift: () => {
        this._cachedMeta.edges.shift();
        this._cachedMeta._parsedEdges.shift();
        this._scheduleResyncLayout();
      },
      onDataSplice: (start, count, ...args) => {
        this._cachedMeta.edges.splice(start, count);
        this._cachedMeta._parsedEdges.splice(start, count);
        if (args.length > 0) {
          const parsed = this._cachedMeta._parsedEdges;
          parsed.splice(start, 0, ...args.map((edge) => this._parseDefinedEdge(edge)));
          this._insertEdgeElements(start, args.length);
        } else {
          this._scheduleResyncLayout();
        }
      },
      onDataUnshift: (...args) => {
        const parsed = this._cachedMeta._parsedEdges;
        parsed.unshift(...args.map((edge) => this._parseDefinedEdge(edge)));
        this._insertEdgeElements(0, args.length);
      },
    };

    // this.resyncLayout();
  }

  parse(start, count) {
    super.parse(start, count);
    // since generated
    this._cachedMeta._sorted = false;
    this._parseEdges();
  }

  reset() {
    if (this._initialReset) {
      this._initialReset = false;
    } else {
      this.resetLayout();
    }
    super.reset();
  }

  update(mode) {
    super.update(mode);

    const meta = this._cachedMeta;
    const edges = meta.edges || [];

    this.updateEdgeElements(edges, 0, mode);
  }

  destroy() {
    super.destroy();
    if (this._edges) {
      unlistenArrayEvents(this._edges, this._edgeListener);
    }
    this.stopLayout();
  }

  updateEdgeElements(edges, start, mode) {
    const bak = {
      _cachedDataOpts: this._cachedDataOpts,
      dataElementType: this.dataElementType,
      dataElementOptions: this.dataElementOptions,
      _sharedOptions: this._sharedOptions,
    };
    this._cachedDataOpts = {};
    this.dataElementType = this.edgeElementType;
    this.dataElementOptions = this.edgeElementOptions;
    const meta = this._cachedMeta;
    const nodes = meta.data;
    const data = meta._parsedEdges;

    // const reset = mode === 'reset';

    const firstOpts = this.resolveDataElementOptions(start, mode);
    const sharedOptions = this.getSharedOptions(mode || 'normal', edges[start], firstOpts);
    const includeOptions = this.includeOptions(mode, sharedOptions);

    for (let i = 0; i < edges.length; i++) {
      const edge = edges[i];
      const index = start + i;
      var parsed = data[index];
      var properties = {
        source: nodes[parsed.source],
        target: nodes[parsed.target],
      };
      if (includeOptions) {
        properties.options = this.resolveDataElementOptions(index, mode);
      }
      this.updateEdgeElement(edge, index, properties, mode);
    }
    this.updateSharedOptions(sharedOptions, mode);

    Object.assign(this, bak);
  }

  updateEdgeElement(edge, index, properties, mode) {
    super.updateElement(edge, index, properties, mode);
  }

  updateElement(point, index, properties, mode) {
    if (mode === 'reset') {
      // start in center also in x
      const xScale = this._cachedMeta.xScale;
      properties.x = xScale.getBasePixel();
    }
    super.updateElement(point, index, properties, mode);
  }

  resolveNodeIndex(nodes, ref) {
    if (typeof ref === 'number') {
      // index
      return ref;
    }
    if (typeof ref === 'string') {
      // label
      const labels = this.chart.data.labels;
      return labels.indexOf(ref);
    }
    const nIndex = nodes.indexOf(ref);
    if (nIndex >= 0) {
      // hit
      return nIndex;
    }

    if (ref && typeof ref.index === 'number') {
      return ref.index;
    }

    const data = this.getDataset().data;
    const index = data.indexOf(ref);
    if (index >= 0) {
      return index;
    }

    console.warn('cannot resolve edge ref', ref);
    return -1;
  }

  buildOrUpdateElements() {
    const dataset = this.getDataset();
    const edges = dataset.edges || [];

    // In order to correctly handle data addition/deletion animation (an thus simulate
    // real-time charts), we need to monitor these data modifications and synchronize
    // the internal meta data accordingly.
    if (this._edges !== edges) {
      if (this._edges) {
        // This case happens when the user replaced the data array instance.
        unlistenArrayEvents(this._edges, this._edgeListener);
      }

      if (edges && Object.isExtensible(edges)) {
        listenArrayEvents(edges, this._edgeListener);
      }
      this._edges = edges;
    }
    super.buildOrUpdateElements();
  }

  draw() {
    const meta = this._cachedMeta;
    const edges = meta.edges || [];
    const elements = meta.data || [];

    const area = this.chart.chartArea;
    const ctx = this._ctx;

    if (edges.length > 0) {
      clipArea(ctx, area);
      edges.forEach((edge) => edge.draw(ctx, area));
      unclipArea(ctx);
    }

    elements.forEach((elem) => elem.draw(ctx, area));
  }

  _resyncElements(changed) {
    super._resyncElements(changed);

    const meta = this._cachedMeta;
    const edges = meta._parsedEdges;
    const metaEdges = meta.edges || (meta.edges = []);
    const numMeta = metaEdges.length;
    const numData = edges.length;

    if (numData < numMeta) {
      metaEdges.splice(numData, numMeta - numData);
      this._scheduleResyncLayout();
    } else if (numData > numMeta) {
      this._insertEdgeElements(numMeta, numData - numMeta);
    } else if (changed) {
      this._scheduleResyncLayout();
    }
  }

  getTreeRootIndex() {
    const ds = this.getDataset();
    const nodes = ds.data;
    if (ds.derivedEdges) {
      // find the one with no parent
      return nodes.findIndex((d) => d.parent == null);
    }
    // find the one with no edge
    const edges = this._cachedMeta._parsedEdges || [];
    const nodeIndices = new Set(nodes.map((_, i) => i));
    edges.forEach((edge) => {
      nodeIndices.delete(edge.targetIndex);
    });
    return Array.from(nodeIndices)[0];
  }

  getTreeRoot() {
    const index = this.getTreeRootIndex();
    const p = this.getParsed(index);
    p.index = index;
    return p;
  }

  getTreeChildren(node) {
    const edges = this._cachedMeta._parsedEdges;
    return edges
      .filter((d) => d.source === node.index)
      .map((d) => {
        const p = this.getParsed(d.target);
        p.index = d.target;
        return p;
      });
  }

  _parseDefinedEdge(edge) {
    const ds = this.getDataset();
    const data = ds.data;
    return {
      source: this.resolveNodeIndex(data, edge.source),
      target: this.resolveNodeIndex(data, edge.target),
    };
  }

  _parseEdges() {
    const ds = this.getDataset();
    const data = ds.data;
    const meta = this._cachedMeta;
    if (ds.edges) {
      return (meta._parsedEdges = ds.edges.map((edge) => this._parseDefinedEdge(edge)));
    }

    const edges = (meta._parsedEdges = []);
    // try to derive edges via parent links
    data.forEach((node, i) => {
      if (node.parent != null) {
        // tree edge
        const parent = this.resolveNodeIndex(data, node.parent);
        edges.push({
          source: parent,
          target: i,
        });
      }
    });
    return edges;
  }

  addElements() {
    super.addElements();

    const meta = this._cachedMeta;
    const edges = this._parseEdges();
    const metaData = (meta.edges = new Array(edges.length));

    for (let i = 0; i < edges.length; ++i) {
      metaData[i] = new this.edgeElementType();
    }
  }

  _resyncEdgeElements() {
    const meta = this._cachedMeta;
    const edges = this._parseEdges();
    const metaData = meta.edges || (meta.edges = []);

    for (let i = 0; i < edges.length; ++i) {
      metaData[i] = metaData[i] || new this.edgeElementType();
    }
    if (edges.length < metaData.length) {
      metaData.splice(edges.length, metaData.length);
    }
  }

  _insertElements(start, count) {
    super._insertElements(start, count);
    if (count > 0) {
      this._resyncEdgeElements();
    }
  }

  _removeElements(start, count) {
    super._removeElements(start, count);
    if (count > 0) {
      this._resyncEdgeElements();
    }
  }

  _insertEdgeElements(start, count) {
    const elements = [];
    for (let i = 0; i < count; i++) {
      elements.push(new this.edgeElementType());
    }
    this._cachedMeta.edges.splice(start, 0, ...elements);
    this.updateEdgeElements(elements, start, 'reset');
    this._scheduleResyncLayout();
  }

  _onDataPush() {
    super._onDataPush.apply(this, Array.from(arguments));
    this._scheduleResyncLayout();
  }
  _onDataPop() {
    super._onDataPop();
    this._scheduleResyncLayout();
  }
  _onDataShift() {
    super._onDataShift();
    this._scheduleResyncLayout();
  }
  _onDataSplice() {
    super._onDataSplice.apply(this, Array.from(arguments));
    this._scheduleResyncLayout();
  }
  _onDataUnshift() {
    super._onDataUnshift.apply(this, Array.from(arguments));
    this._scheduleResyncLayout();
  }

  reLayout() {
    // hook
  }

  resetLayout() {
    // hook
  }

  stopLayout() {
    // hook
  }

  _scheduleResyncLayout() {
    if (this._scheduleResyncLayoutId != null) {
      return;
    }
    this._scheduleResyncLayoutId = requestAnimationFrame(() => {
      this._scheduleResyncLayoutId = null;
      this.resyncLayout();
    });
  }

  resyncLayout() {
    // hook
  }
}

GraphController.id = 'graph';
GraphController.register = () => {
  GraphController.prototype.edgeElementType = EdgeLine.register();
  GraphController.prototype.edgeElementOptions = ScatterController.prototype.datasetElementOptions.concat([
    'tension',
    'stepped',
  ]);
  GraphController.defaults = merge({}, [
    defaults.scatter,
    {
      layout: {
        padding: 5,
      },
      scales: {
        x: {
          display: false,
        },
        y: {
          display: false,
        },
      },
      tooltips: {
        callbacks: {
          label(item, data) {
            return data.labels[item.index];
          },
        },
      },
    },
  ]);
  return registerController(GraphController);
};

class GraphChart extends Chart {
  constructor(item, config) {
    super(item, patchControllerConfig(config, GraphController));
  }
}
GraphChart.id = GraphController.id;

class ForceDirectedGraphController extends GraphController {
  constructor(chart, datasetIndex) {
    super(chart, datasetIndex);
    this._simulation = forceSimulation()
      .on('tick', () => {
        this.chart.update();
      })
      .on('end', () => {
        this.chart.update();
      });
    const sim = chart.options.simulation;

    const fs = {
      center: forceCenter,
      collide: forceCollide,
      link: forceLink,
      manyBody: forceManyBody,
      x: forceX,
      y: forceY,
      radial: forceRadial,
    };

    Object.keys(fs).forEach((key) => {
      const options = sim.forces[key];
      if (!options) {
        return;
      }
      const f = fs[key]();
      if (typeof options !== 'boolean') {
        Object.keys(options).forEach((attr) => {
          f[attr](options[attr]);
        });
      }
      this._simulation.force(key, f);
    });
    this._simulation.stop();
  }

  resetLayout() {
    super.resetLayout();
    this._simulation.stop();

    const nodes = this._cachedMeta._parsed;
    nodes.forEach((node, i) => {
      node.index = i;
      if (!node.reset) {
        return;
      }
      delete node.x;
      delete node.y;
      delete node.vx;
      delete node.vy;
    });
    this._simulation.nodes(nodes);
    this._simulation.alpha(1).restart();
  }

  resyncLayout() {
    super.resyncLayout();
    this._simulation.stop();

    const meta = this._cachedMeta;

    const nodes = meta._parsed;
    nodes.forEach((node, i) => {
      node.index = i;
      if (node.x === null) {
        delete node.x;
      }
      if (node.y === null) {
        delete node.y;
      }
      if (node.x == null && node.y == null) {
        node.reset = true;
      }
    });
    const link = this._simulation.force('link');
    if (link) {
      link.links([]);
    }
    this._simulation.nodes(nodes);
    if (link) {
      // console.assert(ds.edges.length === meta.edges.length);
      // work on copy to avoid change
      link.links((meta._parsedEdges || []).map((link) => Object.assign({}, link)));
    }

    if (this.chart.options.simulation.autoRestart) {
      this._simulation.alpha(1).restart();
    }
  }

  reLayout() {
    this._simulation.alpha(1).restart();
  }

  stopLayout() {
    super.stopLayout();
    this._simulation.stop();
  }
}

ForceDirectedGraphController.id = 'forceDirectedGraph';
ForceDirectedGraphController.register = () => {
  GraphController.register();
  ForceDirectedGraphController.defaults = merge({}, [
    GraphController.defaults,
    {
      simulation: {
        autoRestart: true,
        forces: {
          center: true,
          collide: false,
          link: true,
          manyBody: true,
          x: false,
          y: false,
          radial: false,
        },
      },
    },
  ]);
  return registerController(ForceDirectedGraphController);
};

class ForceDirectedGraphChart extends Chart {
  constructor(item, config) {
    super(item, patchControllerConfig(config, ForceDirectedGraphController));
  }
}
ForceDirectedGraphChart.id = ForceDirectedGraphController.id;

class DendogramController extends GraphController {
  updateEdgeElement(line, index, properties, mode) {
    properties._orientation = this.chart.options.tree.orientation;
    super.updateEdgeElement(line, index, properties, mode);
  }

  updateElement(point, index, properties, mode) {
    properties.angle = this.getParsed(index).angle;
    super.updateElement(point, index, properties, mode);
  }

  resyncLayout() {
    const meta = this._cachedMeta;

    meta.root = hierarchy(this.getTreeRoot(), (d) => this.getTreeChildren(d))
      .count()
      .sort((a, b) => b.height - a.height || b.data.index - a.data.index);

    this.doLayout(meta.root);

    super.resyncLayout();
  }

  reLayout() {
    this.doLayout(this._cachedMeta.root);
  }

  doLayout(root) {
    const options = this.chart.options.tree;

    const layout = options.mode === 'tree' ? tree() : cluster();

    if (options.orientation === 'radial') {
      layout.size([Math.PI * 2, 1]);
    } else {
      layout.size([2, 2]);
    }

    const orientation = {
      horizontal: (d) => {
        d.data.x = d.y - 1;
        d.data.y = -d.x + 1;
      },
      vertical: (d) => {
        d.data.x = d.x - 1;
        d.data.y = -d.y + 1;
      },
      radial: (d) => {
        d.data.x = Math.cos(d.x) * d.y;
        d.data.y = Math.sin(d.x) * d.y;
        d.data.angle = d.y === 0 ? Number.NaN : d.x;
      },
    };

    layout(root).each(orientation[options.orientation] || orientation.horizontal);

    requestAnimationFrame(() => this.chart.update());
  }
}

DendogramController.id = 'dendogram';
DendogramController.register = () => {
  GraphController.register();
  DendogramController.defaults = merge({}, [
    GraphController.defaults,
    {
      tree: {
        mode: 'dendogram', // dendogram, tree
        orientation: 'horizontal', // vertical, horizontal, radial
      },
      datasets: {
        animations: {
          numbers: {
            type: 'number',
            properties: ['x', 'y', 'angle', 'radius', 'rotation', 'borderWidth'],
          },
        },
        tension: 0.4,
      },
      scales: {
        x: {
          min: -1,
          max: 1,
        },
        y: {
          min: -1,
          max: 1,
        },
      },
    },
  ]);
  return registerController(DendogramController);
};

class DendogramChart extends Chart {
  constructor(item, config) {
    super(item, patchControllerConfig(config, DendogramController));
  }
}
DendogramChart.id = DendogramController.id;

class TreeController extends DendogramController {}

TreeController.id = 'tree';
TreeController.register = () => {
  DendogramController.register();
  TreeController.defaults = merge({}, [
    DendogramController.defaults,
    {
      tree: {
        mode: 'tree',
      },
    },
  ]);
  return registerController(TreeController);
};

class TreeChart extends Chart {
  constructor(item, config) {
    super(item, patchControllerConfig(config, TreeController));
  }
}
TreeChart.id = TreeController.id;

export { DendogramChart, DendogramController, EdgeLine, ForceDirectedGraphChart, ForceDirectedGraphController, GraphChart, GraphController, TreeChart, TreeController };
